@@ -9,6 +9,7 @@ import urllib
 import absl
 import tensorflow as tf
 import tensorflow.keras as keras
+import tensorflow_text
 import tfx
 import tensorflow_hub as hub
 import kerastuner as kt
@@ -137,6 +138,7 @@ class TFAutoData():
           features_dict[f_]['Type'] = 'CATEGORICAL'
 
       #Min/Max for numerical features
+      #Count of unique values
       for feat in features_stats['datasets'][0]['features']:
         curr_feat = feat['path']['step'][0]
         if curr_feat == features_dict[f_]['feature'] and features_dict[f_]['Type'] in ['INT','FLOAT']:
@@ -144,6 +146,12 @@ class TFAutoData():
           features_dict[f_]['max'] = feat['numStats'].get('max', 0.0)
           features_dict[f_]['mean'] = feat['numStats'].get('mean', 0.0)
           features_dict[f_]['std_dev'] = feat['numStats'].get('stdDev', 1)
+        elif curr_feat == features_dict[f_]['feature'] and features_dict[f_]['Type'] == 'CATEGORICAL':
+          features_dict[f_]['categorical_values_count'] = {}
+          features_dict[f_]['categorical_values_count_total'] = 0
+          for topvals in feat['stringStats']['topValues']:            
+            features_dict[f_]['categorical_values_count'][topvals.get('value', "NA")] = topvals.get('frequency', 0)
+            features_dict[f_]['categorical_values_count_total'] += topvals.get('frequency', 0)
       
       features_list.append(features_dict[f_])
     self.features_list = features_list
@@ -249,18 +257,23 @@ class TFAutoData():
     return self.pipeline
 
 class TextEncoder(tf.keras.Model):
-  def __init__(self):
-    super(TextEncoder, self).__init__()
-    self.encoder = hub.KerasLayer("https://tfhub.dev/google/universal-sentence-encoder/4", trainable=False)
+  def __init__(self, strategy):
+    self.strategy = strategy
+    with self.strategy.scope():
+      super(TextEncoder, self).__init__()
+      self.preprocessor = hub.KerasLayer("https://tfhub.dev/google/universal-sentence-encoder-cmlm/multilingual-preprocess/2", trainable=False)
+      self.encoder = hub.KerasLayer("https://tfhub.dev/google/LaBSE/2", trainable=False)
 
   def __call__(self, inp):
-    #Preprocess text
-    # encoder_inputs = self.preprocessor(inp)
+    with self.strategy.scope():
+      #Preprocess text
+      # encoder_inputs = self.preprocessor(inp)
 
-    #Encode text
-    embedding = self.encoder(inp)
+      #Encode text
+      preprocess = self.preprocessor(inp)
+      embedding = self.encoder(preprocess)["default"]
 
-    return embedding
+      return embedding
 
 class TFAutoModel():
   def __init__(self, _tfx_root, train_data_path, test_data_path):
@@ -271,7 +284,9 @@ class TFAutoModel():
     self._log_root = os.path.join(self._tfx_root, 'logs');
     self._model_root = os.path.join(self._tfx_root, 'model');
     self._label = ''                                                      #Label
+    self._label_vocab = []
     self._features = []                                                   #List of features to be used for modeling
+    self._class_weights = {}                                              #Class weights
     self._train_data_path = train_data_path                               #Training data
     self._test_data_path = test_data_path                                 #Test data
     self._model_type = ''
@@ -310,6 +325,7 @@ class TFAutoModel():
            'learning_rate':[0.01, 0.005, 0.002, 0.001, 0.0005]
         },
     }
+    self.hpt_config[2] = self.hpt_config[1]
     
     ##GPU Strategy
     self.strategy = tf.distribute.MirroredStrategy()
@@ -374,10 +390,32 @@ class TFAutoModel():
 
       return(features)
 
+    #Convert string labels 0 to N
+    def string_labels_to_num(v_feats, v_label):
+      pass_label = v_label
+      for ix, cat_vals in enumerate(v_feats['categorical_values']):
+        search_string = str(cat_vals) + "$"
+        pass_label = tf.strings.regex_replace(pass_label, search_string, tf.strings.as_string(ix))
+
+      pass_label = tf.strings.to_number(pass_label, out_type=tf.dtypes.int32)
+      return pass_label
+
     #To be called from TF
     def feature_engg(features, label):
       #Add new features
       features = feature_engg_features(features)
+
+      #Replace string label with 0 to N for classification case
+      if self._model_type == 'CLASSIFICATION':
+        for feats in self._config_json['data_schema']:
+          if feats['feature'] == self._label:
+            break
+
+        #Replace categorical values with 0 to N
+        if feats['Type'] == 'CATEGORICAL' and feats['feature'] == self._label:
+          self._label_vocab = feats['categorical_values']
+          label = string_labels_to_num(feats, label)
+
       return(features, label)
 
     def _input_fn(v_test=False):     
@@ -508,16 +546,33 @@ class TFAutoModel():
     return self._feature_cols
 
   def create_keras_model_classification(self, hp):
-    # with self.strategy.scope():
+    with self.strategy.scope():
       # params = self.params_default
       feature_cols = self._feature_cols
 
       #Number of classes
       for feats in self._config_json['data_schema']:
-        #Only include features
+        #Only include label
         if feats['feature'] == self._label:
-          num_classes = int(feats['max'] + 1)
-          break
+          if feats['Type'] != 'CATEGORICAL':
+            num_classes = int(feats['max'] + 1)
+            break
+          else:
+            num_classes = len(feats['categorical_values'])
+            break
+
+      #Calculate class weights
+      for feats in self._config_json['data_schema']:
+        #Only include label
+        if feats['feature'] == self._label:
+          if feats.get('categorical_values', 'NA') != 'NA':
+            #In case of categorical labels
+            for ix, class_ in enumerate(feats['categorical_values']):
+              self._class_weights[ix] = (1 / feats['categorical_values_count'][class_]) * (feats['categorical_values_count_total'])/2.0
+          else:
+            #In case of numerical labels
+            for class_ in range(num_classes):
+              self._class_weights[class_] = 1
 
       METRICS = [
           # tf.keras.metrics.AUC(multi_label=True, num_labels=num_classes),
@@ -552,15 +607,16 @@ class TFAutoModel():
           #Apply normalization
           feat_numeric[-1] = ( tf.cast(feature_cols['K'][feats['feature']], tf.float32) - feats['mean'] ) / feats['std_dev']
 
-      #Handle Text attributes
+      #Handle Text attributes( For model complexity > 2 )
       feat_text = []
-      text_emb = TextEncoder()
-      for feats in self._config_json['data_schema']:
-        if feats['Type'] == 'STRING':
-          feat_text.append('')
-          
-          #Apply Text Encoding from TFHub
-          feat_text[-1] = text_emb(tf.squeeze(tf.cast(feature_cols['K'][feats['feature']], tf.string)))
+      if self._model_complexity >= 2:
+        text_emb = TextEncoder(self.strategy)
+        for feats in self._config_json['data_schema']:
+          if feats['Type'] == 'STRING':
+            feat_text.append('')
+            
+            #Apply Text Encoding from TFHub
+            feat_text[-1] = text_emb(tf.squeeze(tf.cast(feature_cols['K'][feats['feature']], tf.string)))
 
       ###Create MODEL
       ####Concatenate all features( Numerical input )
@@ -626,13 +682,13 @@ class TFAutoModel():
       hp_learning_rate = hp.Choice('learning_rate', values=self.hpt_config[self._model_complexity]['learning_rate'], ordered=True)
       opt = tf.keras.optimizers.Adam(lr = hp_learning_rate)
 
-      #Compile model
-      model.compile(loss='sparse_categorical_crossentropy',  optimizer=opt, metrics = METRICS)
+    #Compile model
+    model.compile(loss='sparse_categorical_crossentropy',  optimizer=opt, metrics = METRICS)
 
-      return model
+    return model
 
   def create_keras_model_regression(self, hp):
-    # with self.strategy.scope():
+    with self.strategy.scope():
       METRICS = [
               keras.metrics.RootMeanSquaredError(name='rmse'),
               keras.metrics.MeanAbsolutePercentageError(name='mape')
@@ -669,15 +725,16 @@ class TFAutoModel():
           #apply normalization
           feat_numeric[-1] = ( tf.cast(feature_cols['K'][feats['feature']], tf.float32) - feats['mean'] ) / feats['std_dev']
       
-      #Handle Text attributes
+      #Handle Text attributes( For model complexity > 2 )
       feat_text = []
-      text_emb = TextEncoder()
-      for feats in self._config_json['data_schema']:
-        if feats['Type'] == 'STRING':
-          feat_text.append('')
-          
-          #Apply Text Encoding from TFHub
-          feat_text[-1] = text_emb(tf.squeeze(tf.cast(feature_cols['K'][feats['feature']], tf.string)))
+      if self._model_complexity >= 2:
+        text_emb = TextEncoder(self.strategy)
+        for feats in self._config_json['data_schema']:
+          if feats['Type'] == 'STRING':
+            feat_text.append('')
+            
+            #Apply Text Encoding from TFHub
+            feat_text[-1] = text_emb(tf.squeeze(tf.cast(feature_cols['K'][feats['feature']], tf.string)))
 
       ###Create MODEL
       ####Concatenate all features( Numerical input )
@@ -742,10 +799,10 @@ class TFAutoModel():
       hp_learning_rate = hp.Choice('learning_rate', values=self.hpt_config[self._model_complexity]['learning_rate'], ordered=True)
       opt = tf.keras.optimizers.Adam(lr = hp_learning_rate)
 
-      #Compile model
-      model.compile(loss='mean_squared_error',  optimizer=opt, metrics = METRICS)
+    #Compile model
+    model.compile(loss='mean_squared_error',  optimizer=opt, metrics = METRICS)
 
-      return model
+    return model
 
   def keras_train_and_evaluate(self, model, epochs=100, mode='Train'):
     #Add callbacks
@@ -781,6 +838,7 @@ class TFAutoModel():
     #                 batch_size = 10)()
     
     #Train and Evaluate
+    #Best model chosen from tuning is just refitted here on full data
     if mode == 'Train':
       if self._model_type == 'REGRESSION':
         print("Training a regression model...")
@@ -794,9 +852,11 @@ class TFAutoModel():
                       # validation_steps = 3,   ###Keep this none for running evaluation on full EVAL data every epoch
                       steps_per_epoch = 100,   ###Has to be passed - Cant help it :) [ Number of batches per epoch ]
                       callbacks=[reduce_lr, #modelsave_callback, #tensorboard_callback, 
-                                keras.callbacks.EarlyStopping(patience=20, restore_best_weights=True, verbose=True)]
+                                keras.callbacks.EarlyStopping(patience=20, restore_best_weights=True, verbose=True)],
+                      class_weight = self._class_weights
                       )
     else:
+      #Model is created during Tuning cycle
       if self._model_type == 'REGRESSION':
         print("Hyper-Tuning a regression model...")
         mod_func = self.create_keras_model_regression
@@ -812,7 +872,8 @@ class TFAutoModel():
                       max_epochs=10)
       stop_early = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=5)
 
-      tuner.search(train_dataset, validation_data=validation_dataset, epochs=50, steps_per_epoch = 100, callbacks=[stop_early])
+      tuner.search(train_dataset, validation_data=validation_dataset, epochs=50, steps_per_epoch = 100, callbacks=[stop_early],
+                   class_weight = self._class_weights)
       # print(f"""
       # The hyperparameter search is complete. The optimal number of units in the first densely-connected
       # layer is {best_hps.get('units')} and the optimal learning rate for the optimizer
@@ -882,21 +943,24 @@ class TFAutoModel():
 
     #Generate examples
     for row in range(max_samples):
-      example = tf.train.Example()
-      #For each column in file
-      for f_ in self._config_json['file_headers']:
-        for feats in self._config_json['data_schema']:
-          if feats['feature'] != f_:
-            continue
+      try:
+        example = tf.train.Example()
+        #For each column in file
+        for f_ in self._config_json['file_headers']:
+          for feats in self._config_json['data_schema']:
+            if feats['feature'] != f_:
+              continue
 
-          #Prepare example data
-          if feats['Type'] in [ 'CATEGORICAL', 'STRING' ]:
-            example.features.feature[f_].bytes_list.value.append(out[f_][row])
-          elif feats['Type'] == 'FLOAT':
-            example.features.feature[f_].float_list.value.append(out[f_][row])
-          elif feats['Type'] == 'INT':
-            example.features.feature[f_].int64_list.value.append(int(out[f_][row]))
-      examples.append(example)
+            #Prepare example data
+            if feats['Type'] in [ 'CATEGORICAL', 'STRING' ]:
+              example.features.feature[f_].bytes_list.value.append(out[f_][row])
+            elif feats['Type'] == 'FLOAT':
+              example.features.feature[f_].float_list.value.append(out[f_][row])
+            elif feats['Type'] == 'INT':
+              example.features.feature[f_].int64_list.value.append(int(out[f_][row]))
+        examples.append(example)
+      except:
+        pass
     return examples
 
   #Create prediction function to link in WIT
@@ -945,12 +1009,15 @@ class TFAutoModel():
     examples_wit = self.generate_examples_for_wit()
     if self._model_type == 'REGRESSION':
       wit_type = 'regression'
+      config_builder = (WitConfigBuilder(examples_wit, self._config_json['file_headers'])
+                  .set_custom_predict_fn(self.wit_prediction_fn_dyn)
+                  .set_model_type(wit_type))
     elif self._model_type == 'CLASSIFICATION':
       wit_type = 'classification'
-
-    config_builder = (WitConfigBuilder(examples_wit, self._config_json['file_headers'])
-                      .set_custom_predict_fn(self.wit_prediction_fn_dyn)
-                      .set_model_type(wit_type))
+      config_builder = (WitConfigBuilder(examples_wit, self._config_json['file_headers'])
+                  .set_custom_predict_fn(self.wit_prediction_fn_dyn)
+                  .set_model_type(wit_type)
+                  .set_label_vocab(self._label_vocab))
     
     WitWidget(config_builder)
   
@@ -962,14 +1029,9 @@ class TFAutoModel():
       if feats['feature'] != self._label:
         continue
 
-      #Only allow numerical values
-      if feats['Type'] in [ 'CATEGORICAL', 'STRING' ]:
-        print("Error: Label should be numerical only")
-        success_flag = False
-        return success_flag
-      
-      if self._model_type == "CLASSIFICATION" and feats['Type'] != 'INT':
-        print("Error: CLASSIFICATION - Label data type is not correct")
+      # Only allow numerical values for REGRESSION models
+      if feats['Type'] in [ 'CATEGORICAL', 'STRING' ] and self._model_type == "REGRESSION":
+        print("Error: REGRESSION - labels should be numerical only")
         success_flag = False
         return success_flag
 
@@ -978,10 +1040,10 @@ class TFAutoModel():
       if feats['feature'] != self._label:
         continue
 
-      #For classification, minimum value should be 0 and type should be INT
-      if self._model_type == "CLASSIFICATION":
+      #For classification, minimum value should be 0 for INT labels
+      if self._model_type == "CLASSIFICATION" and feats['Type'] in ['INT', 'FLOAT']:
         if int(feats['min']) != 0:
-          print("Error: CLASSIFICATION - Label values are not correct")
+          print("Error: CLASSIFICATION - Integer labels should start from 0")
           success_flag = False
           return success_flag
     
@@ -1068,7 +1130,7 @@ class TFAuto():
     Parameters
     label_column: The feature to be used as Label
     model_type: Either of 'REGRESSION', 'CLASSIFICATION'
-    model_complexity: 0 to 1 (0: Model without HPT, 1: Model with HPT) -> More will be added in future
+    model_complexity: 0 to 2 (0: Model without HPT, 1: Model with HPT, 2: Complexity 1 + Also handle Text features)
     '''
     # #Run Modeling steps
     if self.tfadata._run == True:
@@ -1077,7 +1139,7 @@ class TFAuto():
     else:
       print("Error: Please run Step 1 - step_data_explore")
 
-    print("Success: Model Training complete. Exported to {}")
+    print("Success: Model Training complete. Exported to: {}".format(self._model_root + "/"))
 
   def step_model_whatif(self):
     '''
